#include "includevray.h"
#include "../CustomPluginHelpers.h"
using namespace VUtils;

namespace{
	VR::real const pi_over_two = VR::pi( ) * 0.5f;
	VR::real const pi_over_180 = VR::pi( ) / 180.0f;
	VR::real const two_pi = VR::pi() * 2.0f;
}

enum BlendFunction{
	LINEAR,
	NORMAL_DISTRIBUTION,
	SIGMOID,
	GOMPERTZ,
	CUBE_ROOT,
	CUBIC,
	CUBIC_POLYNOMIAL,
};

inline double lerp(double t, double a, double b) {
	return a + t * (b - a);
}

inline float lerp(float t, float a, float b){
	return a + t * (b - a);
}

float cube_root( float x ){
	return powf( x, 0.333333f );
}

//http://en.wikipedia.org/wiki/Gompertz_curve
//http://en.wikipedia.org/wiki/File:Gompertz_com.PNG
float gompertzFunc( float t, float a = 1.0f, float b = 1.0f, float c = 1.0f ){
	return a * expf( b * expf( -c * t ) );
}

// These are helper funcs based on the three graphs in that png - these are helpers because each graph show changes in only one of the a,b, or c vals
float gompertzFunc3( float t, float c = -1.0f ){
	return gompertzFunc(t, 1.0f, -1.0f, c );
}

float gompertzFunc2( float t, float b = 1.0f ){
	return gompertzFunc(t, 1.0f, b );
}

float gompertzFunc1( float t, float a = 1.0f ){
	return gompertzFunc(t,a,-1.0f);
}

// This is a nice S type shape 
//http://en.wikipedia.org/wiki/Sigmoid_function
float sigmaFunction( float t ){
	//P(t) = 1.0f / ( 1 + e ^ -t )
	return 1.0f / ( 1 + expf( -1.0f * t ) );
}

/// http://en.wikipedia.org/wiki/Normal_distribution
/// This will do the pdf for normal distribution - I'm defaulting to the standard where mu is 0 and o^2 is 1
float normalDistributionPdf(float x, float mu=0.0f, float omega_squared = 1.0){
	float const multiplier =  1.0f / ( sqrtf( two_pi * omega_squared ) );

	float const x_minus_mu = (x - mu);

	float const e_exp = ( x_minus_mu * x_minus_mu ) / (-2.0f * omega_squared);
	return multiplier * expf( e_exp );
}

float AngleBetween( VR::Vector const& vec1, VR::Vector const &vec2 ){
	float const dot = static_cast<float>(normalize(vec1) * normalize(vec2));
	if( fabsf( dot ) < 1e-6f ) return pi_over_two;

	return ::acosf( dot );
	//float const dot_prod = static_cast<float>( vec1 * vec2 );
	//// 0 dot product means they are perpendicular
	//if( fabsf( dot_prod ) < 1e-6f ) return( pi_over_two );

	//float const len = vec1.length( ) * vec2.length( );
	//return( ::acos( dot_prod / len ) );
}

/// @fn GetBlendValue
///
/// @brief This function lets us turn a linear blend into an exponential or other function. Thus we can get different blend
/// effects and not just vanilla linear blends.
///
/// @param[in] float blend_ratio - this is the linear blend value ( max-current ) / (max - min )...it tells us how far through the range we are
/// @param[in] BlendFunction blend_function - this will tell us what function we are going to plug the linear blend ratio into 
/// @retval float on the range of 0.0 - 1.0 inclusive
float GetBlendValue( float blend_ratio, BlendFunction function ){
	switch( function ){
		case LINEAR:
			// The value being passed into this function as blend_ratio is the linear blend value...just return that.
			return blend_ratio;
		case NORMAL_DISTRIBUTION:{
			VR::real const val = (VR::real)(lerp(1.0f - blend_ratio,0.0f, 1.0f));
			VR::real pdf_blend = normalDistributionPdf( val );
			
			// i think the standard distribution curve should be 0.4 at its peak
			VR::real const pdf_blend_ratio = 1-pdf_blend; /// 0.4;

			return (val > 0.0f ) ? 1 - pdf_blend_ratio: pdf_blend_ratio;
		}
		case SIGMOID:{
			VR::real const val = (VR::real)(lerp(blend_ratio,-4.0f, 2.0f));
			return sigmaFunction(val);
		}
		case GOMPERTZ:{
			VR::real const val = VR::real(lerp( blend_ratio, -1.5f, 0.0f ));
			//VR::real const blend = gompertzFunc( val, 2 ) / 2.0;
			return gompertzFunc1( val, 2 ) ;/// 2.0f;
		}
		case CUBE_ROOT:{
			VR::real const val = VR::real(lerp( blend_ratio, 0.0f, 4.0f ));
			//VR::real const blend = gompertzFunc( val, 2 ) / 2.0;
			return cube_root(val) / 2.0f;
		}
		case CUBIC:{
			VR::real const val = lerp( blend_ratio, -1.0f, 1.0f );			
			return (powf(2*val, 3) + 1.f);
		}
		case CUBIC_POLYNOMIAL:{
			//on a range of -5 to 5 the min is -100 and the max is 150
			// so I am going to add 100 to make sure our range is 0 - 250
			VR::real const val = lerp( blend_ratio, -5.0f, 5.0f );
			
			// y = x^3 + x^2 + 100 
			// Then I divide by 250 to get it on a scale of 0.0 - 1.0
			return (powf(val,3) + powf(val,2) + 100.0f)  / 250.0f;
		}
	}

	return 0.5f;
}

// Plugin parameters
struct MtlAngleBlendParams : VRayParameterListDesc {
	MtlAngleBlendParams( ) {
		addParamPlugin("mtl_one", EXT_MATERIAL, -1, "the first material to be blended");
		addParamPlugin("mtl_two", EXT_MATERIAL, -1, "the second material to be blended");

		addParamFloat( "start_angle", 25.0f, -1, "angle where blending begins ( degrees ) - anything before this will be material one only" );
		addParamFloat( "stop_angle", 75.0f, -1, "angle where blending ends ( degrees ) - anything after this will be material two only" );

		/*

		LINEAR,
		NORMAL_DISTRIBUTION,
		SIGMOID,
		GOMPERTZ,
		CUBE_ROOT,
		CUBIC,
		CUBIC_POLYNOMIAL,

		 */
		addParamInt( "blend_function", 0, -1, "The function to use to blend the materials. 0:Linear 1:Normal Distribution 2:Sigmoid 3:Gompertz 4:Cube Root 5:Cubic 6: Cubic Polynomial");
		addParamBool( "flip_view_dir", false, -1, "If this is true we will flip the view direction prior to doing calculations on it");
	}
};

class MtlAngleBlend :	public VR::VRayMaterial,
						public RecursiveNewInterfaceProvider{
	public:
		MtlAngleBlend( VRayPluginDesc* desc );
		PluginInterface* newInterface(InterfaceID id) {
			if( EXT_RECURSIVE_NEW_INTERFACE_PROVIDER == id )
				return static_cast<RecursiveNewInterfaceProvider*>(this);

			return VRayMaterial::newInterface(id);
		}
		PluginBase* getPlugin(void) { return (PluginBase*) this; }
	public: // RecursiveNewInterfaceProvider
		PluginInterface* recursiveNewInterface(InterfaceID id) {
			PluginInterface* res = newInterface(id);
			if( res ) return res;

			// we will check to see if either of our inner materials supports this interface...
			if( mtl_one && mtl_one->getPlugin() ){
				res = mtl_one->getPlugin()->newInterface( id );
				if( res )return res;
			}
			if( mtl_two && mtl_two->getPlugin() ){
				res = mtl_two->getPlugin()->newInterface(id);
				if( res )return res;
			}

			return NULL;
		}
	public:// from BSDFInterface
		void frameBegin(VR::VRayRenderer *vray);
		void frameEnd(VR::VRayRenderer *vray);
		void renderBegin( VR::VRayRenderer *vray );
		void renderEnd( VR::VRayRenderer *vray );
		VR::BSDFSampler* newBSDF(VR::VRayContext const &rc, VR::BSDFFlags flags);
		void deleteBSDF( VR::VRayContext const &rc, VR::BSDFSampler *bsdf);
	public: // from MaterialInterface
		/// Shades the surface and leaves the result in rc.mtlresult.
		void shade(VR::VRayContext &rc);
		int getMaterialRenderID( VR::VRayContext const& rc ){ return render_mtl_id; }
		void updateParams(double time=0.0);
	protected:
		MaterialInterface* mtl_one;
		MaterialInterface* mtl_two;
		BSDFInterface* mtl_one_bsdf;
		BSDFInterface* mtl_two_bsdf;
		BRDFPool<LayeredBSDF> bsdf_pool;
		float start_angle, stop_angle;
		int render_mtl_id;
		int blend_function;
		int flip_view_dir;
		int invalidWarning;
};

MtlAngleBlend::MtlAngleBlend( VUtils::VRayPluginDesc* desc )
		: VUtils::VRayMaterial( desc ), mtl_one(0), mtl_two(0){
	paramList->setParamCache( "start_angle", &start_angle );
	paramList->setParamCache( "stop_angle", &stop_angle );
	paramList->setParamCache( "blend_function", &blend_function );
	paramList->setParamCache( "flip_view_dir", &flip_view_dir );

}
void MtlAngleBlend::updateParams(double time/*=0.0*/){
	mtl_one = 0;
	mtl_one_bsdf = 0;

	mtl_two = 0;
	mtl_two_bsdf = 0;

	VRayPluginParameter* mtl_one_param = paramList->getParam( "mtl_one" );
	VRayPluginParameter* mtl_two_param = paramList->getParam( "mtl_two" );

	// get ourselves into radians
	start_angle = start_angle * pi_over_180;
	stop_angle = stop_angle * pi_over_180;

	if( mtl_one_param ){
		mtl_one = static_cast<MaterialInterface*>( GET_INTERFACE( mtl_one_param->getObject( 0, time ), EXT_MATERIAL ) );
		mtl_one_bsdf = static_cast<BSDFInterface*>( GET_INTERFACE( mtl_one_param->getObject(0, time ), EXT_BSDF ) );
	}
	if( mtl_two_param ){
		mtl_two = static_cast<MaterialInterface*>( GET_INTERFACE( mtl_two_param->getObject( 0, time ), EXT_MATERIAL ) );
		mtl_two_bsdf = static_cast<BSDFInterface*>( GET_INTERFACE( mtl_two_param->getObject(0, time ), EXT_BSDF ) );
	}
}
void MtlAngleBlend::frameBegin(VR::VRayRenderer *vray){
	VRayMaterial::frameBegin( vray );
	paramList->cacheParams( );

	render_mtl_id = vray->getRayServer()->getNewRenderID();

	double const time = vray->getFrameData( ).t;
	updateParams(time);

	bsdf_pool.init(vray->getSequenceData().maxRenderThreads);
	invalidWarning=false;
}
void MtlAngleBlend::frameEnd(VR::VRayRenderer *vray){
	VRayMaterial::frameEnd( vray );
	bsdf_pool.freeMem( );
	mtl_one = mtl_two = 0;
	mtl_one_bsdf = mtl_two_bsdf = 0;
}
void MtlAngleBlend::renderBegin( VR::VRayRenderer *vray ){
	VRayMaterial::renderBegin( vray );
	paramList->cacheParams();
	updateParams();
}
void MtlAngleBlend::renderEnd( VR::VRayRenderer *vray ){
	VRayMaterial::renderEnd( vray );
}

VR::BSDFSampler* MtlAngleBlend::newBSDF(VR::VRayContext const &rc, VR::BSDFFlags flags){
	// if there aren't two to blend between we're not going to do anything
	if( !(mtl_one_bsdf && mtl_two_bsdf) )return 0;

	VUtils::Vector effectiveViewDir = rc.rayparams.viewDir;
	if( flip_view_dir )
		effectiveViewDir *= -1.0f;

	// determine angle between view direction and surface normal	
	double const angle_between = AngleBetween( rc.rayresult.normal, effectiveViewDir );
	
	VR::BSDFSampler* res = 0;

	//I want to see what hte angle is in degrees just so its a little easier for me to debug
	double const degree_angle = angle_between * 180.0 / VR::pi();
	double const degree_start_angle = start_angle * 180.0 / VR::pi();
	double const degree_stop_angle = stop_angle * 180.0 / VR::pi();

	if( angle_between < start_angle ){
		res = mtl_one_bsdf->newBSDF( rc, flags );
	}else if( angle_between > stop_angle ){
		res = mtl_two_bsdf->newBSDF( rc, flags );
	}else{
		VRayContext &rcc = const_cast<VR::VRayContext &>( rc );

		LayeredBSDF *layered_bsdf=bsdf_pool.newBRDF(rc);
		layered_bsdf->setNumBSDFs(rcc, 2);
		layered_bsdf->setTransparency(VR::Color(0.0f, 0.0f, 0.0f));

		VR::Color const origMultResult=rc.rayparams.multResult;

		VR::BSDFSampler* mtl_one_new_bsdf = mtl_one_bsdf->newBSDF( rc, flags );
		VR::BSDFSampler* mtl_two_new_bsdf = mtl_two_bsdf->newBSDF( rc, flags );

		VR::real const max_angle = VUtils::Max( start_angle, stop_angle );
		VR::real const min_angle = VUtils::Min( start_angle, stop_angle );
		VR::real const range = max_angle - min_angle;

		/// The page I was looking at online was on the range of -5 - 5....so I need a way to bring the bell curve into my range
		/// or a way to bring my range to that of the bell curve. I guess that is where the mean and standard deviation comes from?

		VR::real const blend_ratio =  (VR::real)(( ( max_angle == min_angle ) ? 0.5f : ( angle_between - min_angle ) / ( range ) ));
		// this will tell me how far between 0 and 5 I am - that was the range they used on the site I got the bellcurve
		// formula from

		// We plug our linear blend value into the chosen blend function to get our final blend value
		float const blend = GetBlendValue( blend_ratio, static_cast<BlendFunction>(blend_function) );
		
		layered_bsdf->setBSDF( 0, mtl_one_new_bsdf, VR::Color( blend, blend, blend ) );
		layered_bsdf->setBSDF( 1, mtl_two_new_bsdf, VR::Color( 0.0f, 0.0f, 0.0f ) );

		rcc.rayparams.multResult=origMultResult;

		res = layered_bsdf;
	}
	return( res );
}
void MtlAngleBlend::deleteBSDF( VR::VRayContext const &rc, VR::BSDFSampler *bsdf){
	// if there aren't two to blend between we're not going to do anything
	if( !(bsdf && mtl_one_bsdf && mtl_two_bsdf) )return;

	VUtils::Vector effectiveViewDir = rc.rayparams.viewDir;
	if( flip_view_dir )
		effectiveViewDir *= -1.0f;

	// determine angle between view direction and surface normal	
	double const angle_between = AngleBetween( rc.rayresult.normal, effectiveViewDir );

	VR::BSDFSampler* res = 0;

	// material one only
	if( angle_between < start_angle ){
		mtl_one_bsdf->deleteBSDF( rc, bsdf );
	// material two only
	}else if( angle_between > stop_angle ){
		mtl_two_bsdf->deleteBSDF( rc, bsdf );
	}else{ // evenly blend both materials
		LayeredBSDF *layered_bsdf= static_cast<LayeredBSDF*>( bsdf );

		VR::BSDFSampler* temp = layered_bsdf->getBSDF( 0 );
		if( temp ) mtl_one_bsdf->deleteBSDF( rc, temp );
		temp = layered_bsdf->getBSDF( 1 );
		if( temp ) mtl_two_bsdf->deleteBSDF( rc, temp );

		bsdf_pool.deleteBRDF( rc, layered_bsdf );
	}

}
void MtlAngleBlend::shade(VR::VRayContext &rc){
	BSDFFlags flags;
	flags.force1sided=false;

	BSDFSampler* bsdf = newBSDF( rc, flags );
	if( !bsdf )return;

	if (rc.rayparams.rayType & RT_SHADOW) {
		rc.mtlresult.color.makeZero();
		rc.mtlresult.transp= bsdf->getBSDFTransparency(rc);
		rc.mtlresult.alpha=rc.mtlresult.transp.whiteComplement();
		rc.mtlresult.alphaTransp=rc.mtlresult.transp;
	} else {
		VR::Color directLight=rc.evalLight(*bsdf);

		// Trace the BRDF forward
		bsdf->traceBSDFForward(rc, true);

		// Add in light contribution
		rc.mtlresult.color+=directLight;
	}

	deleteBSDF( rc, bsdf );

	if (!rc.mtlresult.color.isValid() ) {
		if (!invalidWarning) {
			invalidWarning=true;
			const VRaySequenceData &sdata=rc.vray->getSequenceData();
			if (sdata.progress) sdata.progress->warning("Material \"%s\" produced overbright or invalid color (rgb=%f %f %f).\n", getPluginName(), rc.mtlresult.color.r, rc.mtlresult.color.g, rc.mtlresult.color.b);
		}
		rc.mtlresult.color.makeValid();
	}
}

// Define the plugin
#define MtlAngleBlend_PluginID PluginID(LARGE_CONST(2007110114))
SIMPLE_PLUGIN_LIBRARY(MtlAngleBlend_PluginID, EXT_MATERIAL, "MtlAngleBlend", "A material that blends between two other materials based on angles", MtlAngleBlend, MtlAngleBlendParams );

